#!/usr/bin/env python
# -*- coding: utf-8 -*-

#WRITTEN BY DENE BOWDALO

###------------------------------------------------------------------------------------###
###IMPORT MODULES
###------------------------------------------------------------------------------------###

import __builtin__
import crossnetwork_processing_modules
import datetime
import glob
from mpi4py import MPI
import numpy as np
import os
import pandas as pd
import Queue

###------------------------------------------------------------------------------------###
###------------------------------------------------------------------------------------###

WORKTAG = 1
DIETAG = 0

#LOAD GHOST CONFIGURATION VARIABLES
config_frame = np.load('/esarchive/obs/ghost/process/run/GHOST_configuration.npz')
run_details = config_frame['run_details'].item()
parameter_details = config_frame['parameter_details'].item()
network_details = config_frame['network_details'].item()

#MAKE RUN/PARAMETER/NETWORK DATA GLOBAL ACROSS ALL SUBMODULES
__builtin__.run_details = run_details
__builtin__.parameter_details = parameter_details
__builtin__.network_details = network_details

#CREATE LIST OF ALL MINUTES IN TEMPORAL WINDOW (REPRESENTING START OF TEMPORAL WINDOW), WITH SPACING IN DEFINED TEMPORAL RESOLUTION BETWEEN START_DATE AND END_DATE --> make globally accesible         
start_dt = datetime.datetime(int(run_details['start_year']),int(run_details['start_month']),1,0,0)
end_dt = datetime.datetime(int(run_details['end_year']),int(run_details['end_month']),1,0,0)
all_dt = pd.date_range(start_dt, end_dt, freq=run_details['temporal_resolution_datetime_frequency']).to_pydatetime()
grid_start_minutes = [crossnetwork_processing_modules.datetime_to_fractional_minutes_from_reference(dt) for dt in all_dt]            
__builtin__.grid_start_minutes = grid_start_minutes

#OPTION TO FILTER OUTPUT BY KEY QA CHECKS, SCREENING OUT BAD DATA, BEFORE OUTPUT TO NETCDF
filter_output = False
__builtin__.filter_output = filter_output

class Work(object):
    def __init__(self):
    
        #CREATE QUEUE FOR ADDING DATA TO
        self.work = Queue.Queue()
        
        #GET ALL PROCESSED NATIVE RESOLUTION STATION FILES BY NETWORK/PARAMETER
        station_files = sorted(glob.glob('/esarchive/obs/ghost/%s/native/%s/*.npz'%(run_details['network'],parameter_details['bsc_parameter_name'])))
    
        #MAKE SURE HAVE SOME DATA TO PROCESS
        if len(station_files) > 0:
    
            #TEST FOR PRESENCE OF TEMPORARY GRIDDED STATION DATA FILE DIRECTORY. IF DOES NOT EXIST CREATE IT
            temp_file_dir = '/esarchive/obs/ghost/%s/%s/%s/temporary/'%(run_details['network'],run_details['temporal_resolution_path'],parameter_details['bsc_parameter_name'])
            does_temporary_exist = os.path.isdir(temp_file_dir)
            if does_temporary_exist == False:
                os.makedirs(temp_file_dir)
    
            #TEST FOR PRESENCE OF YEAR/MONTH TEMPORARY DIRECTORIES. IF DO NOT EXIST, CREATE THEM
            process_years = range(int(run_details['start_year']),int(run_details['end_year'])+1)
            for year in process_years:
        
                #is it first process year?
                if year == process_years[0]:
                    months = np.arange(int(run_details['start_month']), 13)
            
                #is it last process year?
                elif year == process_years[-1]:
                    months = np.arange(1,int(run_details['end_month']))
            
                #full year of months otherwise
                else:
                    months = np.arange(1,13)
            
                for month in months:
                    if month < 10:
                        month_string = '0'+str(month)
                    else:
                        month_string = str(month)
                    temp_file_dir = '/esarchive/obs/ghost/%s/%s/%s/temporary/%s/%s/'%(run_details['network'],run_details['temporal_resolution_path'],parameter_details['bsc_parameter_name'],year,month_string)
                    does_temporary_exist = os.path.isdir(temp_file_dir)
                    if does_temporary_exist == False:
                        os.makedirs(temp_file_dir)
                    #DELETE ALL CURRENT TEMPORARY STATION DATA FILES FOR PARAMETER/NETWORK (IN CASE OF FAILED PREVIOUS RUN)
                    for zippath in glob.iglob(os.path.join(temp_file_dir, '*.npz')):
                        os.remove(zippath)

            print 'TEMPORALLY GRIDDING DATA BY STATION: %s'%(run_details['temporal_resolution'])
            #station_files = [station_files[7]]
            #station_files = [station_files[16]]
            #HOW MANY STATIONS TO PROCESS?
            print 'NUMBER OF STATIONS: %s'%(len(station_files))    

            #ADD DATA TO QUEUE
            for station_file in station_files:
                self.work.put(station_file)

    def get_next(self):
        if self.work.empty():
            return None
        return self.work.get()
        
    def test_if_start_empty(self):
        if self.work.empty():
            return True
        else:
            return False
 

def do_work(work):
    crossnetwork_processing_modules.temporally_grid_data(work)
    return work
    
def master(comm):
    num_procs = comm.Get_size()
    status = MPI.Status()
    
    # generate work queue
    wq = Work()
    
    #array for appending results
    results = []
    
    #test if queue is empty to start with
    if wq.test_if_start_empty():
        # if empty - Tell all the slaves to exit by sending an empty message with DIETAG
        for rank in xrange(1, num_procs):
            comm.send(0, dest=rank, tag=DIETAG)
        return

    # Seed the slaves, send one unit of work to each slave (rank)
    working_ranks = []
    for rank in xrange(1, num_procs):
        work = wq.get_next()
        if not work: break
        comm.send(work, dest=rank, tag=WORKTAG)
        working_ranks.append(rank)
    
    # Loop over getting new work requests until there is no more work to be done
    while True:
        work = wq.get_next()
        if not work: break
    
        # Receive results from a slave
        results.append(comm.recv(source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=status))

        # Send the slave a new work unit
        comm.send(work, dest=status.Get_source(), tag=WORKTAG)
    
    # No more work to be done, receive all outstanding results from slaves
    for rank in working_ranks: 
        results.append(comm.recv(source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=status))

    # Tell all the slaves to exit by sending an empty message with DIETAG
    for rank in xrange(1, num_procs):
        comm.send(0, dest=rank, tag=DIETAG)
        
    return results


def slave(comm):
    my_rank = comm.Get_rank()
    status = MPI.Status()

    while True:
        # Receive a message from the master (the work to do)
        work = comm.recv(source=0, tag=MPI.ANY_TAG, status=status)
        # Check the tag of the received message
        if status.Get_tag() == DIETAG: break 

        # Do the work and send it back
        result = do_work(work)
        comm.send(result, dest=0)
        
def main():
    comm = MPI.COMM_WORLD
    my_rank = comm.Get_rank()
    my_name = MPI.Get_processor_name()
    
    if my_rank == 0:
        results = master(comm)
        if results is not None:
            print 'T1 DONE'
        else:
            print 'T1 NO DATA'
    else:
        slave(comm)


if __name__ == '__main__':
    main()